import weakref
import atexit
import asyncio
import sys

from ._logging import get_logger
logger = get_logger()

class _ResourceManager:
    """ This class acts as a single place to manage the destruction of system
    resources (event_loops, ports, etc). To start with, get the singleton
    instance via

    import utils.resource_manager
    """
    def __init__(self):
        # A list of webservers to close when the event loop is closed.
        self.webservers_to_close = []
        self.reg_at_exit(self, 'stop_asyncio_loop')

    def reg_webserver(self, websockets_server):
        """ Register a webserver to be closed when the event loop is stopped. """
        self.webservers_to_close.append(weakref.ref(websockets_server))

    def reg_at_exit(self, obj_or_func=None, fname="", *vargs):
        """ Register a function or method to be called when python exits.
        @param obj_or_func An object containing a method fname, or a global/static function.
        @param fname If obj_or_func is an object instance, then this can be the method name
                     of the object to call.
        @param vargs Any extra parameters to pass to the obj_or_func or fname function.
        """
        if (fname == "" and callable(obj_or_func)):
            atexit.register(obj_or_func, *vargs)
        else:
            ref = weakref.ref(obj_or_func)
            atexit.register(self._call_on_ref, ref, fname, *vargs)

    def stop_asyncio_loop(self):
        """ Calls "stop" on the event_loop and closes any webservers registered
        with reg_webserver. """
        logger.debug("stop_asyncio_loop")
        try:
            event_loop = asyncio.get_running_loop()
            self._close_webservers(event_loop)
            event_loop.stop()
            return
        except RuntimeError:
            self._close_webservers(None)
            return

    def _close_webserver(self, server_weakref, event_loop=None):
        """ Close the given webserver. If event_loop != None, then use the loop
        to wait, blocking until the server has finished closing. """
        instance = server_weakref()
        if (instance == None):
            logger.debug("close_webserver(None)")
            return
        logger.debug(f"close_webserver({instance})")
        instance.close()
        if event_loop != None:
            event_loop.run_until_complete(instance.wait_closed())

    def _close_webservers(self, event_loop):
        """ Calls _close_webserver for each server registered with reg_webserver. """
        for webserver in self.webservers_to_close:
            self._close_webserver(webserver, event_loop)
        self.webservers_to_close.clear()

    def _call_on_ref(self, ref, fname, *vargs):
        """ Call the function named fname on the given object instance ref, aka:

        ref.fname(*vargs)
        """
        instance = ref()
        if instance == None:
            logger.debug('call_on_ref(None)')
            return
        if not hasattr(instance, fname):
            logger.debug(f"call_on_ref({instance.__class__.__name__}.{fname} == None)")
            return
        instance_method = getattr(instance, fname)
        if not callable(instance_method):
            logger.debug(f"call_on_ref(non-callable {instance.__class__.__name__}.{fname})")
            return
        logger.debug(f"{instance.__class__.__name__}.{fname}()")
        instance_method(*vargs)


########################################################################
#
# Copyright (C) 2021,2022
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: casa-feedback@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''cubevis provides a number of python command line tools which can be
used to build GUI applications for astronomy. It also contains some
applications turn-key applications'''

import os as __os


from .private.apps import iclean

try:
    from .__version__ import __version__
except ModuleNotFoundError:
    ###
    ### __version__.py is generated as part of the build, but if the source tree
    ### for cubevis is used directly for development, no __version__.py will be
    ### available so set it to a default value...
    ###
    __version__ = {}


def xml_interface_defs( ):
    '''This function may eventually return XML files for use in generating casashell bindings. An
       indentically named function provided by casatasks allows cubevis to generate an
       interactive clean task interface using the tclean XML file from casatasks.
    '''
    return { }

__mustache_interface_templates__ = { 'iclean': __os.path.join( __os.path.dirname(__file__), "private", "casashell", "iclean.mustache" ) }
def mustache_interface_templates( ):
    '''This provides a list of mustache files provided by cubevis. It may eventually allow
       casashell to generate all of its bindings at startup time. This would allow casashell
       to be consistent with any version of casatasks that is availale.
    '''
    return __mustache_interface_templates__



########################################################################
#
# Copyright (C) 2022, 2024
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: casa-feedback@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''implementation of the ``CreateMask`` application for interactive creation
of channel masks'''

from os.path import exists, splitext, join
from os.path import split as splitpath
import asyncio
from contextlib import asynccontextmanager
from bokeh.layouts import row, column
from bokeh.plotting import show
from bokeh.models import Button, CustomJS, TabPanel, Tabs, Spacer, Div
from cubevis.toolbox import CubeMask, AppContext
from cubevis.bokeh.utils import svg_icon
from bokeh.io import reset_output as reset_bokeh_output
from bokeh.io import output_notebook
from bokeh.models.dom import HTML
from bokeh.models.ui.tooltips import Tooltip
from cubevis.utils import resource_manager, reset_resource_manager, is_notebook
from cubevis.data import casaimage
from cubevis.bokeh.models import TipButton, Tip
from cubevis.utils import ContextMgrChain as CMC

class CreateMask:
    '''Class that can be used to launch a createmask GUI with ``CreateMask('test.im','mask.im')( )``.
    ``CreateMask`` operates in the same manner as ``InteractiveClean``. Regions drawn on the
    displalyed CASA image cube can be added or subtracted from the related but separate mask
    cube where each pixel is a 1 (masked) or a 0 (unmasked).
    '''

    def __stop( self, _=None ):
        if not self.__result_future.done( ):
            self.__result_future.set_result(self.__retrieve_result( ))

    def _abort_handler( self, err ):
        self._error_result = err
        self.__stop( )

    def __reset( self ):
        if self.__initialized:
            reset_bokeh_output( )
            reset_resource_manager( )

        ###
        ### reset asyncio result future
        ###
        self.__result_future = None

        ###
        ### used by data pipe (websocket) initialization function
        ###
        self.__initialized = False

        self._image = None
        ###
        ### error or exception result
        ###
        self._error_result = None

    def __expand_mask_paths( self, path_pairs ):
        '''return expanded mask paths

        Parameters
        ----------
        path_pairs: list of tuples
            each tuple contains a string in the first element that represents an image path
            and the second element is either a None which signals that the an mask path should
            be generated based upon the image path or a string which represents that mask path
            for the first element image.

        Returns
        -------
        [ (str,str) ]
            A list of tuples is returned. The first element of each
            tuple is a string that represents the image path and the second element is a string
            which represents the mask path.
        '''
        uniquified_names = { }        ### dict to keep track of repeated iamges

        def create_mask_path( impath ):
            if impath in uniquified_names and uniquified_names[impath][0] > 1:
                uniq = "%032d" % uniquified_names[impath][1]
                uniquified_names[impath][1] += 1
            else:
                uniq = ''
            path,file = splitpath(impath)
            if len(path) > 0 and not exists(path):
                raise RuntimeError( f'''CreateMask: mask path '{path}' does not exist''' )
            basename,_ = splitext(file)
            return join( path, f'''{basename}{uniq}.mask''')

        ### python zip etc can only be read once
        path_pairs_ = list(path_pairs)
        for p in path_pairs_:
            if p[0] in uniquified_names:
                uniquified_names[p[0]][0] += 1
            else:
                uniquified_names[p[0]] = [1,1]

        return list( map( lambda p: p if p[1] is not None else (p[0],create_mask_path(p[0])), path_pairs_ ) )

    def __create_masks( self, paths ):
        '''Create missing masks...'''
        ### Create any mask images which do not exist (if create=True)
        return list( map( lambda p: (p[0],casaimage.new(*p)) if not exists(p[1]) else p, paths ) )


    def __init__( self, image, mask=None, create=True ):
        '''create a ``createmask`` object which will display image planes from a CASA
        image and allow the user to draw masks for each channel.

        Parameters
        ----------
        image: str or list of str
            path(s) to CASA image for which interactive masks will be drawn

        mask: str or list of str
            path(s) to CASA image mask into which interactive masks will be added.
            None can be used to indicate that the mask should be created
            (if ``create=True``) and the name should be based upon the image name

        create: bool
            if True then the mask will be created if it does not exist
            if False a mask path which does not exist results in an exception
        '''

        ###
        ### Create application context (which includes a temporary directory).
        ### This sets the title of the plot.
        ###
        self._app_state = AppContext( 'Create Mask' )

        ###
        ### widgets shared across image tabs (masking multiple images)
        ###
        self._cube_palette = None
        self._image_bitmask_controls = None

        ###
        ### With Bokeh 3.2.2, the spectrum and convergence plots extend beyond the edge of the
        ### browser window (requiring scrolling) if a width is not specified. It could be that
        ### this should be computed from the width of the tabbed control area at the right of
        ### the image display.
        ###
        self._spect_plot_width = 450

        ###
        ### Validate image paths
        ###
        if isinstance(image, str):
            image_paths = [ image ]
        elif isinstance(image, list) and all(isinstance(x, str) for x in image):
            image_paths = image
        else:
            raise RuntimeError( 'CreateMask: image parameter should be a string or a list of strings' )

        if len(image_paths) == 0:
            raise RuntimeError( 'CreateMask: at least one image path must be specified' )

        for img in image_paths:
            if not exists(img):
                raise RuntimeError(f'''CreateMask: image path '{img}' does not exist''')

        ###
        ### Validate mask paths
        ###
        mask_paths = None
        if mask is None:
            mask_paths = [None] * len(image_paths)
        elif isinstance(mask,str):
            if len(image_paths) != 1:
                raise RuntimeError( 'CreateMask: when specifying masks, one must be provided for each image (even if its value is None)' )
            mask_paths = [mask]
        elif isinstance(mask,list):
            if len(mask) == 0:
                mask_paths = [None] * len(image_paths)
            elif len(image_paths) != len(mask):
                raise RuntimeError( 'CreateMask: when specifying masks, one must be provided for each image (even if its value is None)' )
            elif all(isinstance(x, str) or x is None for x in mask):
                mask_paths = mask
            else:
                raise RuntimeError( 'CreateMask: image parameter should be a string or a list of elements whose type is string or None' )

        ###
        ### Create mask paths if necessary
        ###
        if not create:
            self._paths = self.__expand_mask_paths( zip(image_paths, mask_paths) )
            for _,msk in self._paths:
                if not exists(msk):
                    raise RuntimeError(f'''CreateMask: mask path '{msk}' does not exist (and create=False)''')
        else:
            self._paths = self.__create_masks( self.__expand_mask_paths( zip(image_paths, mask_paths) ) )

        if not all((casaimage.shape(p[0]) == casaimage.shape(p[1])).all( ) for p in self._paths):
            raise RuntimeError('CreateMask: mismatch between image and mask shapes')


        self._fig = { 'help': None, 'status': None }
        self._mask_state = { }
        self._ctrl_state = { }
        initialization_registered = False
        for paths in self._paths:
            _,name = splitpath(paths[0])
            imdetails = self._mask_state[name] = { 'gui': { 'image': {}, 'fig': {},
                                                            'image-adjust': { } } }

            ###
            ### Use CubeMask init_script to set up a 'beforeunload' handler which will signal to
            ### CubeMask that the app is shuting down. It should then send the final results to
            ### Python and then call the provided callback (the Promise resolve function).
            ###
            ### Waiting for this to complete before the browser tab is unloaded requires waiting
            ### for the promise to be resolved within an async function. Creating one to do the
            ### await works. The promise is resolved when the 'done' function call the Promise's
            ### resolve function which is provide to 'done' as its callback function.
            ###
            ### If debugging this, make only a small change before confirming that exit from the
            ### Python asyncio loop continues to work... seems to be fiddly
            ###
            imdetails['gui']['cube'] = CubeMask( paths[0], mask=paths[1],
                                                 init_script = None if initialization_registered else \
                                                               CustomJS( args={ }, code='''
                                                                         window.addEventListener( 'beforeunload',
                                                                                                  (event) => {
                                                                                                      function donePromise( ) {

                                                                                                          return new Promise( (resolve,reject) => {
                                                                                                                                 // call by name does not work here:
                                                                                                                                 //       document._cube_done(cb=resolve)  ???
                                                                                                                                 if ( document._cube_done ) document._cube_done(null,resolve)
                                                                                                                              } )
                                                                                                      }
                                                                                                      ( async () => { await donePromise( ) } )( )
                                                                                                  } )''' ) )

            initialization_registered = True
            imdetails['image-channels'] = imdetails['gui']['cube'].shape( )[3]


            imdetails['gui']['image']['src'] = imdetails['gui']['cube'].js_obj( )
            imdetails['gui']['image']['fig'] = imdetails['gui']['cube'].image( grid=False, height_policy='max', width_policy='max' )

            if self._fig['help'] is None:
                self._fig['help'] = self._ctrl_state['help'] = imdetails['gui']['cube'].help( rows=[
                    '<tr><td><i>stop button</i></td><td>clicking the stop button will close the dialog and control to python</td></tr>' ],
                                                                                              position='right' )
            imdetails['gui']['channel-ctrl'] = imdetails['gui']['cube'].channel_ctrl( )
            imdetails['gui']['cursor-pixel-text'] = imdetails['gui']['cube'].pixel_tracking_text( margin=(-3, 5, 3, 30) )

            self._fig['status'] = imdetails['gui']['status'] = imdetails['gui']['cube'].status_text( "<p>initialization</p>" , width=230, reuse=self._fig['status'] )
            self._image_bitmask_controls = imdetails['gui']['cube'].bitmask_ctrl( reuse=self._image_bitmask_controls, button_type='light' )

            ###
            ### spectrum plot must be disabled during iteration due to "tap to change channel" functionality
            ###
            if imdetails['image-channels'] > 1:
                imdetails['gui']['spectrum'] = imdetails['gui']['cube'].spectrum( orient='vertical', sizing_mode='stretch_height', width=self._spect_plot_width )
                imdetails['gui']['slider'] = imdetails['gui']['cube'].slider( show_value=False, title='', margin=(14,5,5,5), sizing_mode="scale_width" )
                imdetails['gui']['goto'] = imdetails['gui']['cube'].goto( )
            else:
                imdetails['gui']['spectrum'] = None
                imdetails['gui']['slider'] = None
                imdetails['gui']['goto'] = None

        ###
        ### This is used to tell whether the websockets have been initialized, but also to
        ### indicate if __call__ is being called multiple times to allow for resetting Bokeh
        ###
        self.__initialized = False

        ###
        ### the asyncio future that is used to transmit the result from mask creation
        ###
        self.__result_future = None

    def _create_colormap_adjust( self, imdetails ):
        palette = imdetails['gui']['cube'].palette( reuse=self._cube_palette )
        return column( row( Div(text="<div><b>Colormap:</b></div>",margin=(5,2,5,25)), palette ),
                       imdetails['gui']['cube'].colormap_adjust( ), sizing_mode='stretch_both' )


    def _create_control_image_tab( self, imid, imdetails ):
        result = Tabs( tabs= ( [ TabPanel( child=imdetails['gui']['spectrum'],
                                           title='Spectrum' ) ] if imdetails['image-channels'] > 1 else [ ] ) +
                          [ TabPanel( child=self._create_colormap_adjust(imdetails),
                                      title='Colormap' ),
                            TabPanel( child=column( *imdetails['gui']['cube'].statistics( ) ),
                                      title='Statistics' ) ],
                       width=500, sizing_mode='stretch_height', tabs_location='below' )

        if not hasattr(self,'_image_control_tab_groups'):
            self._image_control_tab_groups = { }

        self._image_control_tab_groups[imid] = result
        result.js_on_change( 'active', CustomJS( args={ },
                                                 code='''document._casa_last_control_tab = cb_obj.active''' ) )
        return result

    def _create_image_panel( self, imagetuple ):
        imid, imdetails = imagetuple

        return TabPanel( child=column( row( *imdetails['gui']['channel-ctrl'], imdetails['gui']['cube'].coord_ctrl( ),
                                            *self._image_bitmask_controls,
                                            #Spacer( height=5, height_policy="fixed", sizing_mode="scale_width" ),
                                            imdetails['gui']['cursor-pixel-text'],
                                            row( Spacer( sizing_mode='stretch_width' ),
                                                 imdetails['gui']['cube'].tapedeck( size='20px' ) if imdetails['image-channels'] > 1 else Div( ),
                                                 Spacer( height=5, width=350 ), width_policy='max' ),
                                            width_policy='max' ),
                                       row( imdetails['gui']['image']['fig'],
                                            column( row( imdetails['gui']['goto'],
                                                         imdetails['gui']['slider'],
                                                         width_policy='max' ) if imdetails['image-channels'] > 1 else Div( ),
                                                    self._create_control_image_tab(imid, imdetails), height_policy='max' ),
                                            height_policy='max', width_policy='max' ),
                                       height_policy='max', width_policy='max' ), title=imid )

    def _launch_gui( self ):
        '''create and show GUI
        '''
        self.__initialized = True

        cwidth = 64
        cheight = 40

        self._ctrl_state['stop'] = TipButton( button_type="danger", max_width=cwidth, max_height=cheight, name='stop',
                                              icon=svg_icon(icon_name="iclean-stop", size=18),
                                              tooltip=Tooltip( content=HTML( '''Clicking this button will cause this tab to close and control will return to Python.''' ),
                                                               position='left' ) )

        self._ctrl_state['stop'].js_on_click( CustomJS( args={ },
                                                        code='''if ( confirm( "Are you sure you want to end this mask creation session and close the GUI?" ) ) {
                                                                    if ( document._cube_done ) document._cube_done( )
                                                                }''' ) )



        tab_panels = list( map( self._create_image_panel, self._mask_state.items( ) ) )

        for imid, imdetails in self._mask_state.items( ):
            imdetails['gui']['cube'].connect( )

        image_tabs = Tabs( tabs=tab_panels, tabs_location='below', height_policy='max', width_policy='max' )

        self._fig['layout'] = column(
                                  row( self._fig['help'],
                                       Spacer( height=self._ctrl_state['stop'].height, sizing_mode="scale_width" ),
                                       Div( text="<div><b>status:</b></div>" ),
                                       self._fig['status'],
                                       self._ctrl_state['stop'], sizing_mode="scale_width" ),
                                  row( image_tabs, height_policy='max', width_policy='max' ),
                                  height_policy='max', width_policy='max' )

        ###
        ### Keep track of which image is currently active in document._casa_image_name (which is
        ### initialized in self._js['initialize']). Also, update the current control sub-tab
        ### when the field main-tab is changed. An attempt to manage this all within the
        ### control sub-tabs using a reference to self._image_control_tab_groups from
        ### each control sub-tab failed with:
        ###
        ###     bokeh.core.serialization.SerializationError: circular reference
        ###
        image_tabs.js_on_change( 'active', CustomJS( args=dict( names=[ t[0] for t in self._mask_state.items( ) ],
                                                                itergroups=self._image_control_tab_groups ),
                                                     code='''if ( ! hasprop(document,'_casa_last_control_tab') ) {
                                                                 document._casa_last_control_tab = 0
                                                             }
                                                             document._casa_image_name = names[cb_obj.active]
                                                             itergroups[document._casa_image_name].active = document._casa_last_control_tab''' ) )

        # Change display type depending on runtime environment
        if is_notebook( ):
            output_notebook()
        else:
            ### Directory is created when an HTTP server is running
            ### (MAX)
###         output_file(self._imagename+'_webpage/index.html')
            pass

        show(self._fig['layout'])

    def _asyncio_loop( self ):
        '''return the event loop which can be mixed in with an existing event loop
        to allow GUI and websocket events to be processed.
        '''
        return self._cube.loop( )

    def __call__( self ):
        '''Display GUI using the event loop specified by ``loop``.
        '''
        async def _run_( ):
            async with self.serve( ) as s:
                await s
        asyncio.run(_run_( ))
        return self.result( )

    @asynccontextmanager
    async def serve( self ):
        '''This function is intended for developers who would like to embed interactive
        clean as a part of a larger GUI. This embedded use of interactive clean is not
        currently supported and would require the addition of parameters to this function
        as well as changes to the interactive clean implementation. However, this function
        does expose the ``asyncio.Future`` that is used to signal completion of the
        interactive cleaning operation, and it provides the coroutines which must be
        managed by asyncio to make the interactive clean GUI responsive.
        '''
        self.__reset( )
        self._launch_gui( )

        async with CMC( *( [ ctx for img in self._mask_state.keys( ) for ctx in
                             [
                                 self._mask_state[img]['gui']['cube'].serve(self.__stop),
                             ]
                           ] ) ):
            self.__result_future = asyncio.Future( )
            yield self.__result_future
        ###async with self._cube.serve( self.__stop ) as cube:
        ###    self.__result_future = asyncio.Future( )
        ###    yield ( self.__result_future, { 'cube': cube } )

    def __retrieve_result( self ):
        '''If InteractiveClean had a return value, it would be filled in as part of the
        GUI dialog between Python and JavaScript and this function would return it'''
        if isinstance(self._error_result,Exception):
            raise self._error_result
        if self._error_result is not None:
            return self._error_result
        return self._paths

    def result( self ):
        '''If InteractiveClean had a return value, it would be filled in as part of the
        GUI dialog between Python and JavaScript and this function would return it'''
        if self.__result_future is None:
            raise RuntimeError( 'no interactive clean result is available' )
        return self.__result_future.result( )



# Functions for long-running process running in a remote jupyter kernel

from multiprocessing import Process, Pipe
import sys
import os
import json
from time import sleep
from websockets.server import serve
import asyncio

class TestProc:
    _backchannel_task = None # Maintain strong reference to backchannel task

    def __init__(self, port, loop):
        self.loop = loop
        self.port = port
        self.subproc_cmd_pend = []

    async def echo(self, websocket):
        async for message in websocket:
            response = json.dumps({"request": message, "response": eval("self."+message)})
            await websocket.send(response)

    async def backchannel_process(self, port):
        async with serve(self.echo, "localhost", port, ping_interval=None):
            await asyncio.Future()  # run forever

    def start_backchannel(self):
        assert self.loop.is_running()
        self._backchannel_task = asyncio.create_task(self.backchannel_process(self.port))
        return

    '''Check subprocess command to see if it's free (1) or busy (0)'''
    def subproc_is_free(self):
        if len(self.subproc_cmd_pend) == 0:
            return 1

        if any(process[1].is_alive() for process in self.subproc_cmd_pend):
            print("Busy")
            sleep(1)
            return 0
        else:
            print('All processes done')
            return 1

    def add_subproc_cmd(self, cmd_json_str):
        # Check that a subprocess isn't currently running
        if not self.subproc_is_free():
            print("Error: Subprocess is still running. Cannot start new process.")
            return -1

        print(cmd_json_str)
        cmd_json = json.loads(cmd_json_str)
        parent_conn, child_conn = Pipe()
        p = Process(target=eval(cmd_json['cmd']),
                    args=(child_conn,) + tuple(cmd_json['args']),
                    kwargs=cmd_json['kwargs'])

        subproc_details = (cmd_json['id'], p, parent_conn)
        self.subproc_cmd_pend.append(subproc_details)
        p.start()
        return

    # Return the requested process' (identified by the UUID) return value
    def get_return_val(self, id):
        # Check that id is valid
        for process in self.subproc_cmd_pend:
            if id == process[0]:
                print("placeholder")
                # Check that process is done
                # TODO

                # Check parent connection for waiting message
                # ret_obj = process[3].recv() ???

                # Check that ret_obj valid and that didn't timeout
                # return ret_obj

        # Else return ID not found
        return



########################################################################
#
# Copyright (C) 2025
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: casa-feedback@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''Specialization of the Bokeh ``DataSource`` which allows for update via WebSockets.'''

from uuid import uuid4
from sys import stderr
from contextlib import asynccontextmanager
from bokeh.plotting import ColumnDataSource
from bokeh.core.properties import Instance, String, Nullable
from bokeh.models.callbacks import Callback
from bokeh.models import CustomJS
import websockets
from ._data_pipe import DataPipe
from ..state import casalib_url, cubevisjs_url
from ...utils import find_ws_address

class UpdatableDataSource(ColumnDataSource):
    """Implementation of a `ColumnDataSource` customized to allow updates from Python via
    WebSockets after the Bokeh GUI has been displayed. This class is derived from `ColumnDataSource`
    and provides an `update` function that can be used to update the plotted data. While this
    is useful when Bokeh is used directly in a Python session, the `stream` function provided
    by `ColumnDataSource` should be used to update plots which are implemented using the Bokeh
    server. This function can be used for plots which may displayed and managed directly from an
    interactive Python session via `asyncio`. However, even in this case it may be more performant
    to include all of the initial plot points when creating the data source. Future updates in
    response to user interaction could be accomplished using this class.

    This class allows the use to specify `js_init` code that is run in the browser when this
    class is initialized there. The JavaScript implementation provides default code which is
    run in the browser when the user calls the `update` function from Python. This code works
    for single dimemsion data by simply appending the supplied updates to the end of the existing
    plot data. If the user supplies an implementation for `js_update` then this code is used
    in preference to the simple, default implementation. In the future the default update
    implementation may be expanded to be more useful for 2D plots.

    Attributes
    ----------
    js_init: CustomJS
        JavaScript code to be run when this DataSource is loaded
    js_update: CustomJS
        JavaScript code to be run to update the data
    pipe: DataPipe
        Optional attribute that allows the user to specify a reused DataPipe
    data:
        INHERITED contains the data provided by this DataSource
    """

    js_init = Nullable(Instance(Callback), help="""
    JavaScript to be run during initialization of an instance of this DataSource object.
    """)

    js_update = Nullable(Instance(Callback), help="""
    JavaScript to be run to update during initialization of an instance of this DataSource object.
    The default behavior is to append to the existing data.
    """)

    pipe = Instance(DataPipe, help="""
    Optional attribute that allows for reusing a single DataPipe for multiple purposes
    """ )

    session_id = String( help="""
    Internal id used for communcations.
    """ )

    __javascript__ = [ casalib_url( ), cubevisjs_url( ) ]

    __created_pipe = False
    __id = str(uuid4( ))
    __user_init = None

    def __init__( self, *args,
                  callback=None,
                  **kwargs ):

        ### websocket communication id
        kwargs['session_id'] = self.__id

        if 'pipe' not in kwargs:
            kwargs['pipe'] = DataPipe( address=find_ws_address( ) ) #, abort=abort )
            self.__created_pipe = True
        if 'js_init' in kwargs:
            self.__user_init = kwargs['js_init']

        kwargs['js_init'] = CustomJS( args=dict( userinit=self.__user_init ),
                                      code='''this._data_keys = casalib.reduce(
                                                  (acc,key,value) => {
                                                      if ( Array.isArray(value) ) acc.push( key )
                                                      return acc }, this.data, [ ] )

                                              this.pipe.register( this.session_id,
                                                                  ( msg ) => {
                                                                      // [...array1, ...array2]
                                                                      if ( ! ('data' in msg) )
                                                                          return { result: 'error', msg: 'expected to find "data" in message' }
                                                                      for (const k of this._data_keys) {
                                                                          if ( ! (k in msg.data) )
                                                                              return { result: 'error', msg: `expected "${k}" field in update "data"` }
                                                                      }
                                                                      if ( this.js_update ) {
                                                                          this.js_update.execute( this, { data: msg.data } ).then(
                                                                             (data) => { this.data = data } )
                                                                      } else {
                                                                          const data_update = { }
                                                                          for (const k of this._data_keys) {
                                                                              const existing = this.data[k]
                                                                              const addition = msg.data[k]
                                                                              data_update[k] = [...existing, ...addition]
                                                                          }
                                                                          this.data = data_update
                                                                      }
                                                                      return { result: 'success',
                                                                               msg: casalib.reduce(
                                                                                        (acc,key,value) => {
                                                                                            if ( ! ( key in this.data ) )
                                                                                                acc[key] = value
                                                                                            return acc
                                                                                        }, msg.data, { } ) }
                                                                  } )

                                              window.addEventListener(
                                                  'beforeunload',
                                                  (event) => {
                                                      function send_shutdown(id,pipe) {
                                                          return new Promise((resolve) =>
                                                              pipe.send( id, { action: 'stop' },
                                                                         (msg) => { resolve(msg) } ) )
                                                      }
                                                      send_shutdown(this.session_id,this.pipe).then(
                                                          (result) => { console.log(`Shutdown Result: ${result}`) } )
                                                  } )

                                              if ( userinit != null ) {
                                                  void userinit.execute( this )
                                              }''' )

        super( ).__init__( *args, **kwargs )
        self.__callback = callback
        self.__stop = None

        async def manage_requests( msg, self=self ):
            result = 'OK'
            if self.__stop and 'action' in msg and msg['action'] == 'stop':
                self.__stop( msg )
            elif self.__callback and 'action' in msg and 'message' in msg and msg['action'] == 'callback':
                result = await self.__callback( msg['message'] )
            return { 'result': result }

        self.pipe.register( self.__id, manage_requests )

    async def update( self, data, callback=None ):
        def default_callback( msg, self=self ):
            if msg['result'] == 'error':
                print( f'''UpdatableDataSource Error: {msg['msg']}''', file=stderr )
        await self.pipe.send( self.__id, { 'action': 'append', 'data': data }, default_callback if callback is None else callback )

    @asynccontextmanager
    async def serve( self, stop_function ):
        self.__stop = stop_function
        if not self.__created_pipe:
            raise RuntimeError( 'UpdatableDataSource.serve should only be called for when a "pipe" is NOT supplied as part of initialization' )

        self._stop_serving_function = stop_function
        async with websockets.serve( self.pipe.process_messages, self.pipe.address[0], self.pipe.address[1] ) as msgpipe:
            yield { 'msgpipe': msgpipe }



########################################################################
#
# Copyright (C) 2023
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: casa-feedback@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''Implementation of ``CBResetTool`` which is a version of the
``ResetTool`` which resets a graph and calls a callback to allow
 for resetting other state'''

from bokeh.models import ResetTool, Tool
from bokeh.core.properties import Instance, Nullable
from bokeh.models.callbacks import Callback

class CBResetTool(ResetTool):
    '''Tool that emits press and pressup events to signal start and end of drag'''

    precallback = Nullable(Instance(Callback), help="""
    Callback to be called before resetting the graph.
    """)

    postcallback = Nullable(Instance(Callback), help="""
    Callback to be called after resetting the graph.
    """)

    # explicit __init__ to support Init signatures
    def __init__(self, *args, **kwargs) -> None:
        ### neither svg files nor SVGIcon are supported by the icon parameter to a tool
        super().__init__(*args, **kwargs )

Tool.register_alias("cbreset", lambda: CBResetTool())



import asyncio
import logging
import sys

from functools import wraps

###
### from: https://stackoverflow.com/a/38730848
###
class ContextMgrChain(object):

    def __init__(self, *managers):
        self.managers = managers
        self.stack = []
        self.values = []

    async def push(self, manager):
        try:
            if hasattr(manager, '__aenter__'):
                value = await manager.__aenter__()
            else:
                value = manager.__enter__()

            self.stack.append(manager)
            self.values.append(value)
            return value
        except:
            # if we encounter an exception somewhere along our enters,
            # we'll stop adding to the stack, and pop everything we've
            # added so far, to simulate what would happen when an inner
            # block raised an exception.
            swallow = await self.__aexit__(*sys.exc_info())
            if not swallow:
                raise

    async def __aenter__(self):
        value = None

        for manager in self.managers:
            value = await self.push(manager)

        return value

    async def __aexit__(self, exc_type, exc, tb):
        excChanged = False
        swallow = False # default value
        while self.stack:
            # no matter what the outcome, we want to attempt to call __aexit__ on
            # all context managers
            try:
                swallow = await self._pop(exc_type, exc, tb)
                if swallow:
                    # if we swallow an exception on an inner cm, outer cms would
                    # not receive it at all...
                    exc_type = None
                    exc = None
                    tb = None
            except:
                # if we encounter an exception while exiting, that is the
                # new execption we send upward
                excChanged = True
                (exc_type, exc, tb) = sys.exc_info()
                swallow = False

        if exc is None:
            # when we make it to the end, if exc is None, it was swallowed
            # somewhere along the line, and we've exited everything successfully,
            # so tell python to swallow the exception for real
            return True
        elif excChanged:
            # if the exception has been changed, we need to raise it here
            # because otherwise python will just raise the original exception
            if not swallow:
                raise exc
        else:
            # we have the original exception still, we just let python handle it...
            return swallow

    async def _pop(self, exc_type, exc, tb):
        manager = self.stack.pop()
        if hasattr(manager, '__aexit__'):
            return await manager.__aexit__(exc_type, exc, tb)
        else:
            return manager.__exit__(exc_type, exc, tb)



########################################################################
#
# Copyright (C) 2022
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: casa-feedback@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
"""
plotants module
"""

import os

import numpy as np
from bokeh.io import export_png, export_svgs
from bokeh.models import ColumnDataSource, LabelSet
from bokeh.plotting import figure, show

try:
    from cairosvg import svg2pdf
    _have_svg2pdf = True
except ImportError:
    _have_svg2pdf = False

try:
    import casatools as ct
    from casatools import table, msmetadata, quanta, ms, measures
except:
    ct = None
    from cubevis.utils import warn_import
    warn_import('casatools')

_FIGURE_PLOT_WIDTH = 450
_FIGURE_PLOT_HEIGHT = 450


def __get_observatory_info(msname):
    """Extract the observatory information from `msname`.

    Parameters
    ----------
    msname: string
        Path to a CASA measurement set.

    Returns
    -------
    ( string, list )
        string is the telescope name and the list is the array position
    """
    metadata = ct.msmetadata()
    metadata.open(msname)
    telescope = metadata.observatorynames()[0]
    positions = metadata.observatoryposition()
    metadata.close()
    return telescope, positions


def __get_antenna_info(msname, log, exclude, checkbaselines):
    """Return the antenna position info.

    Parameters
    ----------
    msname: string
        Path to the CASA measurement set.
    log: boolean
        whether to plot logarithmic positions
    exclude: [ int ]
        list antenna name/id selection to exclude from plot
    checkbaselines: boolean
        whether to check baselines in the main table
    """
    if ct is None:
        raise RuntimeError('casatools is not available')

    me = ct.measures()
    qa = ct.quanta()
    tb = ct.table()

    telescope, positions = __get_observatory_info(msname)
    positions_wgs84 = me.measure(positions, "WGS84")
    array_lon, array_lat, = [positions_wgs84[i]["value"] for i in ["m0", "m1", "m2"]]

    # Open the ANTENNA subtable to get the names of the antennas in this MS and
    # their positions.  Note that the entries in the ANTENNA subtable are pretty
    # much in random order, so antenna_names translates between their index and name
    # (e.g., index 11 = STD155).  We'll need these indices for later, since the
    # main data table refers to the antennas by their indices, not names.

    anttabname = msname + "/ANTENNA"
    tb.open(anttabname)
    # Get antenna names from antenna table
    antenna_names = np.array(tb.getcol("NAME")).tolist()
    station_names = np.array(tb.getcol("STATION")).tolist()
    if telescope == "VLBA":  # names = ant@station
        antenna_names = ["@".join(antsta) for antsta in zip(antenna_names, station_names)]
    # Get antenna positions from antenna table
    antenna_positions = np.array(
        [
            me.position("ITRF", qa.quantity(x, "m"), qa.quantity(y, "m"), qa.quantity(z, "m"))
            for (x, y, z) in tb.getcol("POSITION").transpose()
        ]
    )
    tb.close()

    all_ant_ids = range(len(antenna_names))
    if checkbaselines:
        # Get antenna ids from main table; this will add to runtime
        tb.open(msname)
        ants1 = tb.getcol("ANTENNA1")
        ants2 = tb.getcol("ANTENNA2")
        tb.close()
        ant_ids_used = list(set(np.append(ants1, ants2)))
    else:
        # use them all!
        ant_ids_used = all_ant_ids

    # handle exclude -- remove from ant_ids_used
    for ant_id in exclude:
        try:
            ant_name_id = antenna_names[ant_id] + " (id " + str(ant_id) + ")"
            ant_ids_used.remove(ant_id)
            print( f'''Exclude antenna {ant_name_id}''' )
        except ValueError:
            print( f'''Cannot exclude antenna {ant_name_id}: not in main table''' )

    # apply ant_ids_used mask
    antenna_names = [antenna_names[i] for i in ant_ids_used]
    antenna_positions = [antenna_positions[i] for i in ant_ids_used]
    station_names = [station_names[i] for i in ant_ids_used]

    n_ants = len(ant_ids_used)
    # casalog.post("Number of points being plotted: " + str(n_ants))
    if n_ants == 0:  # excluded all antennas
        return telescope, antenna_names, [], [], [], []

    # Get the names, indices, and lat/lon/alt coords of "good" antennas.
    ant_wgs84s = np.array([me.measure(pos, "WGS84") for pos in antenna_positions])

    # Convert from lat, lon, alt to X, Y, Z (unless VLBA)
    # where X is east, Y is north, Z is up,
    # and 0, 0, 0 is the center
    # Note: this conversion is NOT exact, since it doesn't take into account
    # Earth's ellipticity!  But it's close enough.
    if telescope == "VLBA" and not log:
        ant_lons, ant_lats = [[pos[i] for pos in ant_wgs84s] for i in ["m0", "m1"]]
        ant_xs = [qa.convert(lon, "deg")["value"] for lon in ant_lons]
        ant_ys = [qa.convert(lat, "deg")["value"] for lat in ant_lats]
    else:
        ant_lons, ant_lats = [np.array([pos[i]["value"] for pos in ant_wgs84s]) for i in ["m0", "m1"]]
        rade = 6370000.0 # radE
        ant_xs = (ant_lons - array_lon) * rade * np.cos(array_lat)
        ant_ys = (ant_lats - array_lat) * rade
    return telescope, antenna_names, ant_ids_used, ant_xs, ant_ys, station_names


def __plot_antennas_log(telescope, names, ids, xpos, ypos, antindex, stations, title):
    raise NotImplementedError("This is a placeholder for another type of plot. It is not implemented yet.")


def __plot_antennas(telescope, names, ids, xpos, ypos, antindex, stations, title):
    if telescope == "VLBA":
        labelx = "Longitude (deg)"
        labely = "Latitude (deg)"
    else:
        # use m or km units
        units = " (m)"
        if np.median(xpos) > 1e6 or np.median(ypos) > 1e6:
            xpos /= 1e3
            ypos /= 1e3
            units = " (km)"
        labelx = "X" + units
        labely = "Y" + units
    if antindex:
        names = [f"{name} ({idx})" for name, idx in zip(names, ids)]
    source = ColumnDataSource(data=dict(x=[], y=[], labels=[]))
    source.data = dict(x=xpos, y=ypos, labels=names)
    labels = LabelSet(x="x", y="y", text="labels", x_offset=5, y_offset=5, source=source, text_font_size="10pt")
    plot = figure(plot_height=_FIGURE_PLOT_HEIGHT, plot_width=_FIGURE_PLOT_WIDTH)
    plot.scatter("x", "y", source=source, size=5, line_color="red", fill_color="red", fill_alpha=0.5)
    plot.title.text = title
    plot.xaxis[0].axis_label = labelx
    plot.yaxis[0].axis_label = labely
    plot.add_layout(labels)
    return plot


def plotants(vis, figfile="", antindex=False, logpos=False, exclude=[], checkbaselines=False, title="", showgui=True):
    """Plot the antenna distribution in the local reference frame:
    The location of the antennas in the MS will be plotted with
    X-toward local east; Y-toward local north.  The name of each
    antenna is shown next to its respective location.

    Parameters
    ----------
    vis: string
    Path to the input visibility file

    antindex: boolean, default: False
        Label antennas with name and antenna ID

    logpos: boolean, default: False
        Produce a logarithmic position plot.

    exclude: list, default: []
        antenna IDs or names to exclude from plotting, for example:
        exclude=[2,3,4], exclude=['DV15']

    checkbaselines: boolean, default: False
        Only plot antennas in the MAIN table. This can be useful after a split.
        WARNING:  Setting checkbaselines to True will add to runtime in proportion
        to the number of rows in the dataset.

    title: string, default: ''
        Title written along top of plot
    """
    if os.path.exists(vis) is False:
        raise Exception(f"Visibility file {vis} does not exist")  # could be a print + return
    # remove trailing / for title basename
    if vis.endswith("/"):
        vis = vis[:-1]

    myms = ct.ms()
    try:
        exclude = myms.msseltoindex(vis, baseline=exclude)["antenna1"].tolist()
    except RuntimeError as rterr:  # MSSelection failed
        errmsg = str(rterr)
        errmsg = errmsg.replace("specificion", "specification")
        errmsg = errmsg.replace("Antenna Expression: ", "")
        raise RuntimeError("Exclude selection error: " + errmsg) from rterr

    # Get the antenna positions
    telescope, names, ids, xpos, ypos, stations = __get_antenna_info(vis, logpos, exclude, checkbaselines)
    if not names:
        raise ValueError("No antennas selected. Exiting plotants.")

    if title == "":
        msname = os.path.basename(vis)
        title = "Antenna Positions for "
        if len(msname) > 55:
            title += "\n"
        title += msname

    if logpos:
        fig = __plot_antennas_log(telescope, names, ids, xpos, ypos, antindex, stations, title)
    else:
        fig = __plot_antennas(telescope, names, ids, xpos, ypos, antindex, stations, title)

    if showgui is not False:
        show(fig)

    if figfile != "":
        if figfile.endswith(".png"):
            export_png(fig, filename=figfile)
        elif figfile.endswith(".svg"):
            fig.output_backend = "svg"
            export_svgs(fig, filename=figfile)
        elif figfile.endswith(".pdf"):
            if _have_svg2pdf == True:
                fig.output_backend = "svg"
                export_svgs(fig, filename=figfile.replace(".pdf", ".svg"))
                svg2pdf(url=figfile.replace(".pdf", ".svg"), write_to=figfile)
                os.system("rm " + figfile.replace(".pdf", ".svg"))
            else:
                raise RuntimeError("cairosvg is required for generating PDF output, but it is not available")
        else:
            raise ValueError("Invalid output file type.  Must be .png or .svg or .pdf")



########################################################################
#
# Copyright (C) 2022,2024,2025
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: casa-feedback@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''Common tools used in creating applications.'''

from ._cube import CubeMask
from ._app_context import AppContext
from ._region_list import RegionList



import os
import ssl
import certifi
import urllib
import tarfile

from os.path import splitext
from os.path import split as splitpath
from cubevis import iclean

from argparse import ArgumentParser
name = splitext(splitpath(__file__)[1])[0]

argparse = ArgumentParser( prog=name, description='Basic tests of iclean' )
argparse.add_argument( '--twopoint', action='store_true',
                       help='use refim_twopoints_twochan.ms instead of refim_point_withline.ms' )
args = argparse.parse_args( )

if args.twopoint:
    img = 'test2pt'
    ms_path = 'refim_twopoints_twochan.ms'
    ms_url = "https://casa.nrao.edu/download/devel/casavis/data/refim_twopoints_twochan-ms.tar.gz"
    cleanargs = dict( cell='8.0arcsec', phasecenter="J2000 19:59:28.500 +40.44.01.50",
                      deconvolver='hogbom', specmode='cube' )
else:
    img = 'test'
    ms_path = 'refim_point_withline.ms'
    ms_url = "https://casa.nrao.edu/download/devel/casavis/data/refim_point_withline-ms.tar.gz"
    cleanargs = dict( cell='12.0arcsec', specmode='cube', interpolation='nearest',
                      nchan=5, start='1.0GHz', width='0.2GHz', pblimit=-1e-05,
                      deconvolver='hogbom', threshold='0.001Jy', cyclefactor=3, scales=[0,3,10] )

if not os.path.isdir(ms_path):
    try:
        context = ssl.create_default_context(cafile=certifi.where())
        tstream = urllib.request.urlopen(ms_url, context=context, timeout=400)
        tar = tarfile.open(fileobj=tstream, mode="r:gz")
        tar.extractall( )
    except urllib.error.URLError:
        print("Failed to open connection to "+ms_url)
        raise

if not os.path.isdir(ms_path):
    raise  RuntimeError("Failed to fetch measurement set")

print("Result = " + str( repr( iclean( vis=ms_path, imagename=img,
                                       imsize=512, niter=50,
                                       cycleniter=10,
                                       **cleanargs ) ) ) )



import os
import ssl
import certifi
import urllib
import tarfile

import asyncio
import time
import numpy as np

from bokeh.plotting import figure, show
from bokeh.models import CustomJS
from casatools import ms as mstool
from cubevis.utils import serialize, deserialize, find_ws_address
from cubevis.bokeh.sources import UpdatableDataSource

ms_path = 'tl018b.ms'
ms_url = 'https://casa.nrao.edu/download/devel/casavis/data/tl018b_ms.tar.gz'

if not os.path.isdir(ms_path):
    try:
        context = ssl.create_default_context(cafile=certifi.where())
        tstream = urllib.request.urlopen(ms_url, context=context, timeout=400)
        tar = tarfile.open(fileobj=tstream, mode="r:gz")
        tar.extractall()
    except urllib.error.URLError:
        print("Failed to open connection to "+ms_url)
        raise

ms1 = mstool( )
ms1.open(ms_path)
ms1.msselect({'scan': "91",'baseline':"OV,PT"})

send_count = 0

data1 = None
data2 = None
domain = None
offset = 0
update_size = 1024 * 30

started_up_future = None
async def event_handler( msg ):
    global started_up_future
    print( '>>>ms-update>>event_handler>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>' )
    print( msg )
    started_up_future.set_result(True)
    return { 'result': "OK" }

source = UpdatableDataSource( dict(x=[], y1=[], y2=[]),
                              callback=event_handler,
                              js_init=CustomJS( args={},
                                                code='''this.send( {msg: 'STARTED'}, (msg) => { console.log('Recieved', msg) } )''' ),
                             )

# Create a plot
p = figure()
p.scatter('x', 'y1', source=source, color='blue')
p.scatter('x', 'y2', source=source, color='red')

show(p)

async def update_function( ):
    global data1, data2, domain, offset, update_size, send_count
    print(f"In update function at {time.strftime('%X')}")
    off = min(domain.size,offset+update_size)
    send_count = send_count + 1
    await source.update( dict( x=domain[offset:off], y1=data1[offset:off], y2=data2[offset:off], id=send_count ),
                         lambda reply: print( f'''\tcomplete: {reply}''' ) )
    offset = offset + update_size
    return True if offset >= domain.size else False

async def transmit(func, *args, **kwargs):
    global started_up_future, data1, data2, domain, send_count, update_size
    started_up_future = asyncio.Future( )
    await asyncio.sleep(0.5)
    data = ms1.getdata(['DATA'])
    data1 = np.abs(data['data'][0].flatten( ))
    data2 = np.abs(data['data'][1].flatten( ))
    domain = np.array(range(0,min(data1.size,data2.size)))
    await started_up_future
    stop = False
    while stop is False:
        await asyncio.sleep(0.01)
        stop = await func(*args, **kwargs)
        print( f'''Plot size:\t{send_count*update_size} (x2)''' )


async def main( ):
    global domain
    future = asyncio.Future( )
    def stop_function( result ):
        print( 'stop function', result )
        future.set_result( result )
    async with source.serve( stop_function ) as upsrc:
        task = asyncio.create_task(transmit(update_function))
        await future
        task.cancel( )

if __name__ == "__main__":
    asyncio.run(main())



########################################################################
#
# Copyright (C) 2021,2022,2023
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: casa-feedback@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''This contains functions to inject the ``cubevisjs`` library into the
generated HTML that is used to display the Bokeh plots that ``casagui``'s
applications produce'''
import os
import re
from os import path
from os.path import dirname, join, basename, abspath
from urllib.parse import urlparse
from bokeh import resources
from ...utils import path_to_url, static_vars, have_network


@static_vars( initialized=False,
              do_local_subst=not have_network( ) )
def initialize_bokeh( bokehjs_subst=None ):
    """Initialize `bokeh` for use with the ``cubevisjs`` extensions.

    The ``cubevisjs`` extensions for Bokeh are built into a stand-alone
    external JavaScript file constructed for the specific version of
    Bokeh that is being used. Previously the CASA minified JavaScript
    libraries were loaded from the network. Now these libraries are
    included in a ``__js__`` directory within the cubevis Python package.
    If users need to debug JavaScript problems, non-minified versions
    can be copied into the ``__js__`` directory to replace the existing
    libraries. To debug within the bokehjs library, a URL or a path
    can be supplied by the ``bokehjs_subst`` parameter and it will be
    substituted for the standard Bokeh library.

    Parameters
    ----------
    bokehjs_subst: None or str or list of str
        Bokeh dependent javascript library which is loaded after all
        other Bokeh libraries have been loaded. This path could be a
        local path, a URL or None. None is the default in which case
        it loads the published library for the current version of
        ``cubevisjs`` and ``bokeh``

    Example
    -------
    from cubevis.bokeh.state import initialize_bokeh
    initialize_bokeh( bokehjs_subst="/tmp/bokeh-3.2.2.js" )
    """

    if initialize_bokeh.initialized:
        ### only initialize once...
        return

    initialize_bokeh.initialized = True
    resources.Resources._old_js_files = resources.Resources.js_files
    def js_files( self ):
        ########################################################################
        ### Function that will replace the member function within Bokeh that ###
        ### returns the URLs for Bokeh initialization...                     ###
        ########################################################################
        def expand_paths( replacement ):
            ### turn all paths into URLs
            ### turn a single string into a list with one string
            if replacement is None:
                return [ ]
            if type(replacement) == str:
                if path.isfile(replacement):
                    return [ f'''file://{abspath(replacement)}''' ]
                elif replacement.startswith('http'):
                    return [ replacement ]
                else:
                    raise RuntimeError( f'''debugging bokehjs substitution ('{replacement}') does not exist''' )
            if type(replacement) == list:
                result = [ ]
                for u in replacement:
                    if path.isfile(u):
                        result.append( f'''file://{abspath(u)}''' )
                    elif u.startswith('http'):
                        result.append( u )
                    else:
                        raise RuntimeError( f'''debugging bokehjs substitution ('{u}') does not exist''' )
                return result
            return [ ]
        def cubevisjs_predicate( s ):
            ### detect cubevisjs library URL
            return basename(s).startswith('cubevisjs')
        def replace_bokehjs( urls, replacement ):
            ### substitute replacement list for the bokehjs library URL
            result = [ ]
            for url in urls:
                if re.match( r'.*/bokeh-\d+\.\d+\.\d+(?:\.min)?\.js$', url ):
                    result += replacement
                else:
                    result.append(url)
            return result

        def include_all_bokehjs( urls ):
            result = urls.copy( )
            bokeh_path=None
            has_bokeh_widgets=False
            has_bokeh_tables=False
            for url in urls:
                if re.match( r'.*/bokeh-\d+\.\d+\.\d+(?:\.min)?\.js$', url ):
                    bokeh_path = url
                if re.match( r'.*/bokeh-widgets-\d+\.\d+\.\d+(?:\.min)?\.js$', url ):
                    has_bokeh_widgets = True
                if re.match( r'.*/bokeh-tables-\d+\.\d+\.\d+(?:\.min)?\.js$', url ):
                    has_bokeh_tables = True

            if bokeh_path:
                if not has_bokeh_widgets:
                    result.append( bokeh_path.replace('/bokeh-','/bokeh-widgets-') )
                if not has_bokeh_tables:
                    result.append( bokeh_path.replace('/bokeh-','/bokeh-tables-') )

            return result

        user_bokehjs_replacement = expand_paths(bokehjs_subst)
        sys_urls = include_all_bokehjs( resources.Resources._old_js_files.fget(self) )
        if initialize_bokeh.do_local_subst:
            local_urls = [ ]
            for url in sys_urls:
                pieces = urlparse(url)
                if pieces.scheme != 'file':
                    lib = basename(pieces.path)
                    url = path_to_url(lib)
                    if lib != url:
                        local_urls.append(url)
                    else:
                        raise RuntimeError( '''Cannot confirm internet access and could not find local cached versions of Bokeh's libraries''' )
                else:
                    local_urls.append(url)
            sys_urls = local_urls

        if len(user_bokehjs_replacement) == 0:
            ### move cubevisjs library after Bokeh libraries
            return [x for x in sys_urls if not cubevisjs_predicate(x)] + [x for x in sys_urls if cubevisjs_predicate(x)]
        else:
            ### replace bokehjs library with bokehjs_subst and move cubevisjs library after Bokeh libraries
            return replace_bokehjs( [x for x in sys_urls if not cubevisjs_predicate(x)], user_bokehjs_replacement ) + [x for x in sys_urls if cubevisjs_predicate(x)]

    resources.Resources.js_files = property(js_files)
    return



########################################################################
#
# Copyright (C) 2021,2022,2023
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: casa-feedback@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''General utility functions used by the ``cubevis`` tools and applications.'''

import inspect
import itertools
import urllib.request
import urllib.error
from os.path import dirname, join
import sys

from itertools import groupby, chain
from socket import socket
from os import path as __path
from ._ResourceManager import _ResourceManager
from ._logging import get_logger
from ._regions import polygon_indexes
from ._docenum import DocEnum
from ._copydoc import copydoc
from ._pkgs import find_pkg, load_pkg

from astropy import units
from regions import PixCoord
from regions import RectanglePixelRegion, PolygonPixelRegion

try:
    from casatools import regionmanager

    __have_casatools = True
except ImportError:
    __have_casatools = False

logger = get_logger()

from ._conversion import pack_arrays
from ._conversion import strip_arrays
from ._conversion import serialize, deserialize
from ._static import static_vars, static_dir
from ._tiles import TMSTiles
from ._contextmgrchain import ContextMgrChain
from ._import_protected_module import ImportProtectedModule

@static_vars(mgr=None)
def resource_manager( ):
    if resource_manager.mgr is None:
        resource_manager.mgr = _ResourceManager( )
    return resource_manager.mgr

def reset_resource_manager( ):
    if resource_manager.mgr is not None:
        resource_manager.mgr = _ResourceManager( )

def path_to_url(path):
    '''Convert a single filesystem path to a URL.

    If the string specified in the ``path`` parameter exists. It is turned into a
    fully qualified path and converted to a URL and returned. If ``path`` does not
    exist, ``path`` is returned unchanged.

    Parameters
    ----------
    path: str
        path to be checked and expanded

    Returns
    -------
    str
        ``path`` converted to a URL if ``path`` exists, otherwise ``path`` unchanged
    '''
    if __path.exists(path):
        return "file://" + __path.abspath(path)
    else:
        jslib_path = join(dirname(dirname(__file__)),'__js__',path)
        if __path.exists(jslib_path):
            return "file://" + __path.abspath(jslib_path)
        else:
            return path


def find_ws_address(address='127.0.0.1'):
    '''Find free port on ``address`` network and return a tuple with ``address`` and port number

    This function uses the low level socket function to find a free port and return
    a tuple representing the address plus port number.

    Parameters
    ----------
    address: str
        network to be probed for an available port

    Returns
    -------
    tuple of str and int
        network address (`str`) and port number (`int`)
    '''
    sock = socket()
    sock.bind((address, 0))
    result = sock.getsockname()
    sock.close()
    return result


def partition(pred, iterable):
    '''Split ``iterable`` into two lists based on ``pred`` predicate.
    '''
    trues = []
    falses = []
    for item in iterable:
        if pred(item):
            trues.append(item)
        else:
            falses.append(item)
    return trues, falses


def error_msg(*args, **kwargs):
    '''standard method for reporting errors which do not result in aborting out of python

    This function takes the standard set of arguments that the python ``print`` function takes.
    The primary difference is that the output will go to ``stderr`` and perhaps other error
    logs.
    '''
    print(*args, file=sys.stderr, **kwargs)


@static_vars(msgs=dict(
    casatools='{package} is not available so interactive clean and plotants will not work',
    casatasks='{package} is not available so interactive clean will not work'
), reported={})
def warn_import(package):
    '''standard method for reporting (optional) package import failure

    Parameters
    ----------
    package: str
        name of a package whose attempted import failed
    '''
    if package not in warn_import.reported:
        warn_import.reported[package] = True
        error_msg("warning, %s" % warn_import.msgs[package].format(package=package))


@static_vars(url='http://clients3.google.com/generate_204')
def have_network():
    '''check to see if an active network with general internet connectivity
    is available. returns ``True`` if we have internet connectivity and
    ``False`` if we do not.
    '''
    ###
    ### see: https://stackoverflow.com/questions/50558000/test-internet-connection-for-python3
    ###
    try:
        with urllib.request.urlopen(have_network.url) as response:
            return response.status == 204
    except urllib.error.HTTPError:
        ### http error
        return False
    except urllib.error.ContentTooShortError:
        return False
    except urllib.error.URLError:
        return False
    except Exception:
        return False


def ranges(iterable, order=sorted, key=lambda x: x):
    '''collect elements of ``iterable`` into tuple ranges where each tuple represents
    a concesecutive range within the iterable. ``key`` can be used to provide ranges
    for other objects where ``key(element)`` returns the key to be used for sorting
    into ranges'''
    for a, b in groupby(enumerate(
            order(iterable, key=key) if 'key' in inspect.getfullargspec(order).kwonlyargs else order(iterable)),
                        lambda pair: key(pair[1]) - pair[0]):
        b = list(b)
        yield b[0][1], b[-1][1]


def contiguous_ranges(iterable, order=sorted, key=lambda x: x):
    '''split iterable into contiguous index sequence, i.e. reduce consecutive index runs
    to a tuple containing the first and last. ``iterable`` can be a sequence of values that
    can be subtracted or a list of more complex values where the function ``key`` returns
    the key to be used for ordering. ``iterable`` is ordered by ``order`` which by default
    sorts ``iterable``.'''
    unique = list(set(iterable))
    for a, b in groupby(
            enumerate(order(unique, key=key) if 'key' in inspect.getfullargspec(order).kwonlyargs else order(unique)),
            lambda pair: key(pair[1]) - pair[0]):
        b = list(b)
        yield b[0][1], b[-1][1]


def expand_range_incl(r):
    '''expand the tuple supplied as ``r`` into a range which *includes* the first
    and last element ``r``'''
    if r[0] == r[1]:
        return [0]
    else:
        return range(r[0], r[1] + 1)


def __get_wcs_object(csys: 'image.coordsys', naxis=2) -> 'image.coordsys()':
    """Generate world coordinate object from coordinate system.

    Returns:
        w: world coordinate object
    """
    try:
        import numpy as np
        from astropy.wcs import WCS

        w = WCS(naxis=naxis)

        rad_to_deg = 180/np.pi
        w.wcs.crpix = csys.referencepixel()['numeric'][0:2]
        w.wcs.cdelt = csys.increment()['numeric'][0:2]*rad_to_deg
        w.wcs.crval = csys.referencevalue()['numeric'][0:2] * rad_to_deg
        w.wcs.ctype = ['RA---SIN', 'DEC--SIN']

        return w

    except ImportError as error:
        print('Error importing module:: ' + str(error))

    except Exception as error:
        print(error)

def __index_to_stokes(index: int):
    """Convert stokes axis index to alphabetic value.

    Args:
        index (int): enumerated index defining stokes value.

    Returns:
        str: String indicating stokes value. 
    """
    STOKES_MAP = ['I', 'Q', 'U', 'V']
  
    try:
        return [STOKES_MAP[i] for i in index]
    except TypeError:
        return STOKES_MAP[index]

def index_to_stokes(index: int):
    """Convert stokes axis index to alphabetic value.

    Args:
        index (int): enumerated index defining stokes value.

    Returns:
        str: String indicating stokes value. 
    """
    STOKES_MAP = ['I', 'Q', 'U', 'V']
  
    try:
        return [STOKES_MAP[i] for i in index]
    except TypeError:
        return STOKES_MAP[index]

def __get_center_pixels(params: dict):
    """Get center pixel offset value from dictionary.

    Args:
        params (dict): Submask from mask grouping used by itertools

    Returns:
        list: Pixel offset defined in submask.
    """
    return params['d']

def __write_casa_region(region_object: 'astropy.region', coord:str, polygon_shape: str)->list:
    ''' Convert astropy region to casa region string. 

    Parameters
    ----------
    region_objects: astropy region
        Region definition in pixel or world coordinates
    coord: str
        Coordinate system that should be used in the returned masks. Allowed values are 'pixel', 'world'.
    polygon_shape: str
        Region shape definition: rect: centerbox/rotbox, poly:poly 

    '''
    
    meta_value = []
    for key, value in region_object.meta.items():
        if key == 'corr':
            meta_value.append('{key}=[{pol}]'.format(key=key, pol=value))
        elif key == 'range':
            meta_value.append('{key}=[{lower}chan, {upper}chan]'.format(key=key, lower=str(value[0]), upper=str(value[1])))
        else:
            print('Unknown key: {} skipping'.format(key))
        
    meta = ', '.join(meta_value)

    if polygon_shape=='rect':
        if coord=='pixel':
            return 'centerbox[[{x} pix, {y} pix], [{width} pix, {height} pix]], {meta}'.format(
                x=region_object.center.x,
                y=region_object.center.y,
                width=region_object.width,
                height=region_object.height,
                meta=meta
            )
  
        if coord=='world':
            return 'rotbox[[{ra}, {dec}], [{width}, {height}], {angle}], {meta}'.format(
                ra=region_object.center.ra,
                dec=region_object.center.dec,
                width=region_object.width,
                height=region_object.height,
                angle=region_object.angle,
                meta=meta
            )
  
        else:
            raise RuntimeError('Unknown coordinate value: {}'.format(coord))
  
    if polygon_shape=='poly':    
        if coord=='pixel':
            coords_array = []
            for value in region_object.vertices:
                coords_array.append('[{x} pix, {y} pix]'.format(x=value.x, y=value.y))

            coords = ', '.join(coords_array)

            return 'poly[{coords}], {meta}'.format(coords=coords, meta=meta)

        if coord=='world':
            coords_array = []
            for value in region_object.vertices:
                coords_array.append('[{ra}, {dec}]'.format(ra=value.ra, dec=value.dec))

            coords = ', '.join(coords_array)

            return 'poly[{coords}], {meta}'.format(coords=coords, meta=meta)
        else:
            raise RuntimeError('Unknown coordinate value: {}'.format(coord))
    else:
        raise RuntimeError('Unknown polygon shape: {}'.format(polygon_shape))

def convert_masks(masks: dict, coord='pixel', cdesc=None)->list:
    '''Convert masks in standard format (as defined by ``CubeMask.jsmask_to_raw``) into
    other formats like list of CRTF, single region, etc.

    Parameters
    ----------
    masks: dict
        Dictionary containing ``masks`` and ``polys`` keys. The values for ``masks`` are the polygon
        references for each channel and the values for ``polys`` contain the points making up each
        polygon.
    coord: str
        Coordinate system that should be used in the returned masks. Allowed values are 'pixel'.
    cdesc: dict
        Dictionary containing ``csys`` and ``shape`` which describes the coordinate system to be
        used for creating world coordinate coordinates. The ``shape`` is required along with the
        coordinate system for coordinate conversion.
    '''

    if cdesc is None or 'csys' not in cdesc or 'shape' not in cdesc:
        raise RuntimeError('region operations requires a coordinate description (cdesc parameter)')
    else:
        if not isinstance( cdesc['shape'], tuple ):
            raise RuntimeError("coordinate description must contain a 'shape' element of type 'tuple'")

        if isinstance( cdesc['csys'], dict ):
            csys = cdesc['csys']


    full_mask_params = []

    # Get the wcs object of the request format is world
    if coord == 'world':
        wcs = __get_wcs_object(cdesc['csys'])

    # Build list containing masks with channel and polaization information
    # in a more accesible way for the grouping funciton
    for index, mask in masks['masks'].items():
        for sub_mask in mask:
            sub_mask['s'], sub_mask['c'] = index
            full_mask_params.append(sub_mask)

    region_list = []

    # Group masks by unique center pixel and determine channel range. Extract 
    # mask properties and build astropy region for each center pixel and geometry
    # including region meta data.
    for key, props in itertools.groupby(full_mask_params, __get_center_pixels):
        channel_range = []
        for prop in props:
            channel_range.append(prop['c'])
            poly = prop['p']
            center_pixels = prop['d']
      
            mask_shape = masks['polys'][poly]['type']
      
            xs = masks['polys'][poly]['geometry']['xs']
            ys = masks['polys'][poly]['geometry']['ys']

            width = max(xs) - min(xs)
            height = max(ys) - min(ys)

            stokes_index = prop['s']
    
        if mask_shape=='rect':
            region = RectanglePixelRegion(
                PixCoord(x=center_pixels[0], y=center_pixels[1]), 
                width=width, height=height, angle=0*units.deg
        )
        elif mask_shape=='poly':
            region = PolygonPixelRegion(
                vertices=PixCoord(x=xs, y=ys)
        )

        else:
            raise RuntimeError('Invalid mask shape: {func}-->mask={mask}'.format(func=__convert_masks.__qualname__, mask=mask_shape))

        if coord=='world':
            region = region.to_sky(wcs)
            region.meta = {
                'corr': __index_to_stokes(stokes_index),
                'range': [min(channel_range), max(channel_range)]
            }

            region_list.append(__write_casa_region(region, coord=coord, polygon_shape=mask_shape))
        if coord=='pixel':

            region.meta = {
                'corr': __index_to_stokes(stokes_index),
                'range': [min(channel_range), max(channel_range)]
            }

            region_list.append(__write_casa_region(region, coord=coord, polygon_shape=mask_shape))

    return region_list

def __convert_masks(masks, format='crtf', coord='pixel', ret_type='str', cdesc=None):
    '''Convert masks in standard format (as defined by ``CubeMask.jsmask_to_raw``) into
    other formats like list of CRTF, single region, etc.

    Parameters
    ----------
    masks: dict
        Dictionary containing ``masks`` and ``polys`` keys. The values for ``masks`` are the polygon
        references for each channel and the values for ``polys`` contain the points making up each
        polygon.
    format: str
        Format string that indicates the format that should be returned. Allowed values are
        'region' or 'crtf'.
    coord: str
        Coordinate system that should be used in the returned masks. Allowed values are 'pixel'.
    ret_type: str
        Data structure that should be returned the allowed values are 'str', 'list', 'singleton'
    cdesc: dict
        Dictionary containing ``csys`` and ``shape`` which describes the coordinate system to be
        used for creating world coordinate coordinates. The ``shape`` is required along with the
        coordinate system for coordinate conversion.
    '''
    if format == 'region':
        if __have_casatools == False:
            raise RuntimeError('casatools is not available, cannot create CASA regions')
        if cdesc is None or 'csys' not in cdesc or 'shape' not in cdesc:
            raise RuntimeError('region operations requires a coordinate description (cdesc parameter)')
        else:
            if not isinstance(cdesc['shape'], tuple):
                raise RuntimeError("coordinate description must contain a 'shape' element of type 'tuple'")
            if isinstance(cdesc['csys'], dict):
                csys_dict = cdesc['csys']
            else:
                csys_dict = cdesc['csys'].torecord()

    unique_polygons = {}
    ###
    ### collect masks ordered by (poly_index,dx,dy)
    ###
    for index, mask in masks['masks'].items():
        for poly in mask:
            unique_poly = tuple([poly['p']] + poly['d'])
            if unique_poly in unique_polygons:
                unique_polygons[unique_poly] += [index]
            else:
                unique_polygons[unique_poly] = [index]

    polygon_definitions = {index: poly for index, poly in masks['polys'].items()}

    if format == 'region' and coord == 'pixel':
        rg = regionmanager()
        rg.setcoordinates(csys_dict)

        ###
        ### create a region list given the index of the polygon used, the x/y translation, and the
        ### channels the region should be found on
        ###
        def create_regions_rgn(poly_index, xlate, channels):
            def create_result(shape, points):
                return [rg.fromtext(
                    f"{shape}[{points}],range=[{chan_range[0]}chan,{chan_range[1]}chan],corr=[{','.join(index_to_stokes(expand_range_incl(stokes_range)))}]" ,
                    shape=list(cdesc['shape']))
                        for chan_range in contiguous_ranges([c[1] for c in channels]) for stokes_range in
                        contiguous_ranges([c[0] for c in channels])]

            if polygon_definitions[poly_index]['type'] == 'poly':
                poly_pts = ','.join([f"[{x + xlate[0]}pix,{y + xlate[1]}pix]"
                                     for x in polygon_definitions[poly_index]['geometry']['xs']
                                     for y in polygon_definitions[poly_index]['geometry']['ys']])
                return create_result('poly', poly_pts)
            if polygon_definitions[poly_index]['type'] == 'rect':
                xs = polygon_definitions[poly_index]['geometry']['xs']
                ys = polygon_definitions[poly_index]['geometry']['ys']
                box_pts = f"[{min(xs) + xlate[0]}pix,{min(ys) + xlate[1]}pix],[{max(xs) + xlate[0]}pix,{max(ys) + xlate[1]}pix]"
                return create_result('box', box_pts)

        ### flatten list of unique polygon regions
        result = list(chain.from_iterable(
            [create_regions_rgn(index[0], index[1:], channels) for index, channels in unique_polygons.items()]))
        if ret_type == 'singleton':
            if len(result) == 0:
                raise RuntimeError("no regions created")
            if len(result) == 1:
                return result[0]
            ###return dict(enumerate(result)))          ## see: CAS-13764
            return rg.makeunion(dict(map(lambda t: (str(t[0]), t[1]), list(enumerate(result)))))
        if ret_type == 'list':
            return result
        raise RuntimeError(f"unknown ret_type for region format ({ret_type})")

    if format == 'crtf' and coord == 'pixel':
        ###
        ### create a region list given the index of the polygon used, the x/y translation, and the
        ### channels the region should be found on
        ###
        def create_regions_crt(poly_index, xlate, channels):
            def create_result(shape, points):
                return [
                    f"{shape}[{points}],range=[{chan_range[0]}chan,{chan_range[1]}chan],corr=[{','.join(index_to_stokes(expand_range_incl(stokes_range)))}]"
                    for chan_range in contiguous_ranges([c[1] for c in channels]) for stokes_range in
                    contiguous_ranges([c[0] for c in channels])]

            if polygon_definitions[poly_index]['type'] == 'poly':
                poly_pts = ','.join([f"[{x + xlate[0]}pix,{y + xlate[1]}pix]"
                                     for x in polygon_definitions[poly_index]['geometry']['xs']
                                     for y in polygon_definitions[poly_index]['geometry']['ys']])
                return create_result('poly', poly_pts)
            if polygon_definitions[poly_index]['type'] == 'rect':
                xs = polygon_definitions[poly_index]['geometry']['xs']
                ys = polygon_definitions[poly_index]['geometry']['ys']
                box_pts = f"[{min(xs) + xlate[0]}pix,{min(ys) + xlate[1]}pix],[{max(xs) + xlate[0]}pix,{max(ys) + xlate[1]}pix]"
                return create_result('box', box_pts)

        ###
        ### generate CFTF
        ###
        result = [create_regions_crt(index[0], index[1:], channels) for index, channels in unique_polygons.items()]
        if ret_type == 'str':
            return '\n'.join(chain.from_iterable(result))
        if ret_type == 'list':
            return list(chain.from_iterable(result))
        raise RuntimeError(f"unknown ret_type for crtf format ({ret_type})")
    raise RuntimeError(f"invalid format ({format}), or coord ({coord})")


def set_attributes(obj, **kw):
    '''Given an object and a set of keyword arguments, set the attributes
    in the object that correspond to the keywords to the specified values.

    Parameters
    ----------
    obj: object
        Object whose attributes should be set
    kw: keyword and object
        Attributes to be set

    Returns
    -------
    object
        ``obj`` parameter
    '''
    for k, v in kw.items():
        if hasattr(obj,k): setattr(obj, k, v)
    return obj


def is_notebook() -> bool:
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            if get_ipython().__class__.__module__ == 'google.colab._shell':
                return True   # Google Colab
            else:
              return False  # Other type (?)
    except NameError:
        return False
